#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include <ctime>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <mpi.h>
#include "CNF_mpi.h"

SEQUENCE::SEQUENCE(int length_seq, bCNF_Model* pModel)
{
	m_pModel = pModel;
	this->length_seq = length_seq;
	forward = new ScoreMatrix(m_pModel->num_states,length_seq);
	backward = new ScoreMatrix(m_pModel->num_states,length_seq);
	obs_label = new int[length_seq];
	int df = m_pModel->dim_features*length_seq;
	
	_features = new Score[df];

	predicted_label = new int[length_seq];	
	obs_feature = new Score*[length_seq];
	for(int i=0;i<length_seq;i++)
		obs_feature[i] = new Score[m_pModel->dim_one_pos];
}


SEQUENCE::~SEQUENCE(){
	delete forward;
	delete backward;
	delete obs_label;
	delete _features;

	for(int i=0;i<length_seq;i++)
		delete obs_feature[i];
	delete obs_feature;
}

Score SEQUENCE::Obj_p(){ // predicted label : log-likelihood
	ComputeForward();
	ComputeBackward();
	CalcPartition();
	Score obj = -Partition;
	
	for(int t=0;t<length_seq;t++){
		int leftState = DUMMY;
		if(t>0) leftState = predicted_label[t-1];
		int currState = predicted_label[t];
		obj+=ComputeScore(leftState,currState,t);
	}
	cout << " predicted obj = " << obj << endl;
	return obj;
}

Score SEQUENCE::Obj(){ // observed label : log-likelihood`
	ComputeGates();
	ComputeVi();
	ComputeForward();
	//ComputeBackward();
	CalcPartition();
	Score obj = -Partition;

	for(int t=0;t<length_seq;t++){
		int leftState = GetObsState(t-1);
		int currState = GetObsState(t);
		obj+=ComputeScore(leftState,currState,t);
	}
	
	return obj;
}

void SEQUENCE::ComputeViterbi()
{
	// Viterbi Matrix
	ScoreMatrix best(m_pModel->num_states,length_seq);
	best.Fill((Score)LogScore_ZERO);
	// TraceBack Matrix
	ScoreMatrix traceback(m_pModel->num_states,length_seq);
	traceback.Fill(DUMMY);
	// compute the scores for the first position
	for(int i=0;i<m_pModel->num_states;i++){
		best(i,0)=ComputeScore(DUMMY,i,0);
	}

	// Compute the Viterbi Matrix
	for(int t=1;t<length_seq;t++){
		for(int currState=0;currState<m_pModel->num_states;currState++){
			for(int leftState=0;leftState<m_pModel->num_states;leftState++){
				Score new_score = ComputeScore(leftState,currState,t) + best(leftState,t-1);
				if(new_score > best(currState,t)){
					best(currState,t) = new_score;
					traceback(currState,t) = leftState;
				}
			}
		}
	}
	
	Score max_s = LogScore_ZERO;
	int last_state = 0;
	for(int i=0;i<m_pModel->num_states;i++)
		if(best(i,length_seq-1)>max_s)
			max_s = best(i,length_seq-1), last_state = i;
	//TraceBack
	for(int t=length_seq-1; t>=0;t--){
		predicted_label[t]=last_state;
		last_state=(int)traceback(last_state,t);
	}
}

void SEQUENCE::MAP(){ // Posterior Decoding (Marginal Probability Decoder)
	ComputeForward();
	ComputeBackward();
	for(int t=0;t<length_seq;t++){
		int idx = 0;
		Score maxS = LogScore_ZERO;
		for(int i=0;i<m_pModel->num_states;i++){
			Score s = 0;
			//if(t==0) 
			//	s = (*backward)(i,0); 
			//else 
				s = (*backward)(i,t) + (*forward)(i,t);
			if(s > maxS) maxS = s,idx = i;
		}
		predicted_label[t]=idx;
	}
}

void SEQUENCE::ComputeForward()
{
	forward->Fill(LogScore_ZERO);
	for(int i=0;i<m_pModel->num_states;i++){
		(*forward)(i,0)=ComputeScore(DUMMY,i,0);
	}

	for(int t=1;t<length_seq;t++){
		for(int currState=0;currState<m_pModel->num_states;currState++){
			for(int leftState=0;leftState<m_pModel->num_states;leftState++) {
				Score new_score = ComputeScore(leftState,currState,t);
				LogScore_PLUS_EQUALS((*forward)(currState,t),new_score + (*forward)(leftState,t-1));
			}
		}
	}
}

void SEQUENCE::ComputeBackward()
{
	backward->Fill(LogScore_ZERO);
	for(int i=0;i<m_pModel->num_states;i++){
		(*backward)(i,length_seq-1)=0;
	}

	for(int t=length_seq-2;t>=0;t--){
		for(int currState=0;currState<m_pModel->num_states;currState++){
			for(int rightState=0;rightState<m_pModel->num_states;rightState++){
				Score new_score = ComputeScore(currState,rightState,t+1);
				LogScore_PLUS_EQUALS((*backward)(currState,t),new_score + (*backward)(rightState,t+1));
			}
		}
	}
}

void SEQUENCE::ComputeGates()
{
	gates.resize(m_pModel->num_gates, length_seq);
	for (int pos=0;pos<length_seq; pos++) {
		Score* features=getFeatures(pos);
		for (int k=0;k<m_pModel->num_gates;k++) {
			gates(k,pos) = m_pModel->GetGateOutput(k,features);
		}
	}
}

void SEQUENCE::ComputeVi()
{
	arrVi.resize(m_pModel->num_states, length_seq);
	int num_states = m_pModel->num_states;
	int num_gates = m_pModel->num_gates;
	for (int pos=0;pos<length_seq; pos++)
		for (int currState=0;currState<m_pModel->num_states;currState++) {
			int weightLStart = num_states + num_states*num_states + currState*num_gates;
			Score score = 0;
			for(int k=0;k<m_pModel->num_gates;k++) {
				Score output = gates(k,pos);
				score += m_pModel->weights[weightLStart++]*output;
			}
			arrVi(currState, pos) = score;
		}
}

Score SEQUENCE::ComputeScore(int leftState, int currState, int pos){
	//compute num_gates neurons' output
	Score score = m_pModel->GetWeightT(leftState,currState)+arrVi(currState, pos);
	return score;
}

void SEQUENCE::makeFeatures(){
	//build features for local windows	
	for(int t=0;t<length_seq;t++){
		int pivot = t*m_pModel->dim_features;
		for(int i=0; i< m_pModel->window_size; i++){
			int offset = t+i-m_pModel->window_size/2;
			if(offset <0 || offset >=length_seq){
				for(int j=0;j<m_pModel->dim_one_pos;j++)
					_features[pivot] = 0, pivot++;
			} else {
				for(int j=0;j<m_pModel->dim_one_pos;j++)
					_features[pivot] = obs_feature[offset][j], pivot++;
			}
		}
		_features[pivot] = 1; // last one for the bias term
	}
}

void SEQUENCE::CalcPartition(){
	Partition = (Score)LogScore_ZERO;
	Score FP = (Score)LogScore_ZERO;
	//Score BP = (Score)LogScore_ZERO;
	for(int k=0;k<m_pModel->num_states;k++)
		LogScore_PLUS_EQUALS(Partition, (*forward)(k,length_seq-1));
	//for(int k=0;k<m_pModel->num_states;k++)
	//	LogScore_PLUS_EQUALS(BP, (*backward)(k,0) + (*forward)(k,0));
}


Score* SEQUENCE::getFeatures(int pos){
	int offset;
	offset = pos*m_pModel->dim_features;
	return _features+offset;
}

int SEQUENCE::GetObsState(int pos){
	if(pos<0 || pos>=length_seq) return DUMMY;
	return obs_label[pos];
}

void SEQUENCE::ComputeGradient(bool bCalculateGate){
	int num_states = m_pModel->num_states;
	int num_gates = m_pModel->num_gates; 
	int gstart = num_states*(num_states+1);
	int fstart = gstart + num_states*num_gates;
	double* prob_weight_sum=new double[num_gates];
	// Compute backward forward and partition function;

	struct timeval ts, te;
	gettimeofday(&ts, 0);
	if (bCalculateGate) {
		ComputeGates();
		ComputeVi();
	}
	ComputeForward();
if (proc_id==0) {
	gettimeofday(&te, 0);
	double dif = difftime (te.tv_sec,ts.tv_sec);
	//cerr << "ComputeForward: " << dif*1000+te.tv_usec/1000-ts.tv_usec/1000 << " miliseconds." << endl;
	gettimeofday(&ts, 0);
}
	ComputeBackward();
if (proc_id==0) {
	gettimeofday(&te, 0);
	double dif = difftime (te.tv_sec,ts.tv_sec);
	//cerr << "ComputeBackward: " << dif*1000+te.tv_usec/1000-ts.tv_usec/1000 << " miliseconds." << endl;
	gettimeofday(&ts, 0);
}
	CalcPartition();
if (proc_id==0) {
	gettimeofday(&te, 0);
	double dif = difftime (te.tv_sec,ts.tv_sec);
//	cerr << "CalcPartition: " << dif*1000+te.tv_usec/1000-ts.tv_usec/1000 << " miliseconds." << endl;
	gettimeofday(&ts, 0);
}

	// compute the empirical values
	for(int t=0;t<length_seq;t++){
		int leftState = GetObsState(t-1);
		int currState = GetObsState(t);
		Score* features=getFeatures(t);

		// Trans Weight
		m_pModel->grad[num_states + leftState*num_states + currState] += 1;

		for(int i=0;i<num_gates;i++){
			Score out = gates(i,t);
			// label-based gates weights
			m_pModel->grad[gstart + currState*num_gates + i] += out;

			// Gate weights
			double weight_out = m_pModel->weights[gstart + currState*num_gates + i]*(1.-out)*out;
			int gradStart = fstart + i*m_pModel->dim_features;
			for(int j=0;j<m_pModel->dim_features;j++){
				m_pModel->grad[gradStart + j] += weight_out*features[j];
			}
		}
	}
if (proc_id==0) {
	gettimeofday(&te, 0);
	double dif = difftime (te.tv_sec,ts.tv_sec);
	//cerr << "compute the empirical values: " << dif*1000+te.tv_usec/1000-ts.tv_usec/1000 << " miliseconds." << endl;
	gettimeofday(&ts, 0);
}

	// compute the expected values
	for(int t=0;t<length_seq;t++) {
		Score* features=getFeatures(t);
		memset(prob_weight_sum,0, num_gates*sizeof(double));
		if(t==0) {
			int leftState = DUMMY;
			for(int currState=0;currState<num_states;currState++) {
				Score prob = (*backward)(currState,t) + ComputeScore(leftState,currState,t) - Partition;
				prob = exp(prob);
				// Trans feauture;
				m_pModel->grad[num_states + leftState*num_states + currState] -= prob;

				for(int i=0;i<num_gates;i++){
					// label-based gates weights
					Score out=gates(i,t); 

					int weightIdx = (num_states+1)*num_states + currState*num_gates +i;
					double prob_out = prob*out;
					m_pModel->grad[weightIdx] -= prob_out;

					//inner parameters
					double prob_weight = prob_out*m_pModel->weights[weightIdx]*(1.-out);
					prob_weight_sum[i] +=prob_weight;
				}
			}
		} else {
			// Trans Weights
			for(int leftState=0;leftState<num_states;leftState++)
			   for(int currState=0;currState<num_states;currState++){
				Score prob = (*forward)(leftState,t-1) + (*backward)(currState,t) + ComputeScore(leftState,currState,t) - Partition;
				prob = exp(prob);
				// Trans feauture;
				m_pModel->grad[num_states + leftState*num_states + currState] -= prob;
			   }
			for(int currState=0;currState<num_states;currState++){
				Score prob = (*backward)(currState,t) + (*forward)(currState,t) - Partition;
				prob = exp(prob);

				for(int i=0;i<num_gates;i++){
					// label-based gates weights 
					Score out=gates(i,t); 

					int weightIdx = gstart + currState*num_gates +i;
					double prob_out = prob*out;
					m_pModel->grad[weightIdx] -= prob_out;

					//inner parameters
					double prob_weight = prob_out*m_pModel->weights[weightIdx]*(1.-out);
					prob_weight_sum[i] +=prob_weight;
				}
			}
		}
		for(int i=0;i<num_gates;i++){
			int gradStart = fstart + i*m_pModel->dim_features;
			for(int j=0;j<m_pModel->dim_features;j++){
				m_pModel->grad[gradStart + j] -= prob_weight_sum[i]*features[j];
			}
		}	
	}
	delete prob_weight_sum;
if (proc_id==0) {
	gettimeofday(&te, 0);
	double dif = difftime (te.tv_sec,ts.tv_sec);
	//cerr << "compute the expected values: " << dif*1000+te.tv_usec/1000-ts.tv_usec/1000 << " miliseconds." << endl;
	gettimeofday(&ts, 0);
}
}

void SEQUENCE::ComputeTestAccuracy(){
	m_pModel->totalPos += length_seq;
	// comparison
	for(int t=0; t<length_seq;t++)
		if(obs_label[t]==predicted_label[t])
			m_pModel->totalCorrect++, m_pModel->apw+=1.0/length_seq;
//	for(int t=0; t<length_seq;t++) 
//		cout << predicted_label[t] << ":" << obs_label[t] << " ";
//	cout << endl;
}

void _LBFGS::Report(const vector<double> &theta, int iteration, double objective, double step_length)
{
	//if(iteration%5) return;
	for(int i=0;i<m_pModel->num_params;i++)
		m_pModel->weights[i] = theta[i];
	
	int tc_sum = 0, tp_sum = 0;
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Bcast(m_pModel->weights, m_pModel->num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);

	m_pModel->totalPos = m_pModel->totalCorrect = 0;
	for(int i=0;i<m_pModel->num_tst;i++){
		m_pModel->testData[i]->ComputeGates();
		m_pModel->testData[i]->ComputeVi();
		m_pModel->testData[i]->ComputeViterbi();
		m_pModel->testData[i]->ComputeTestAccuracy();
	}
	double norm_w = 0;
	for(int i=0;i<theta.size();i++)
		norm_w +=theta[i]*theta[i];
	tp_sum = m_pModel->totalPos;
	tc_sum = m_pModel->totalCorrect;
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalPos, &tp_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalCorrect, &tc_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
		
if(proc_id==0){	
	cout << endl << "Iteration:  " << iteration << endl;
	cout << " Weight Norm: " << sqrt(norm_w) << endl;
	cout << " Objective: " << objective << endl;
	cout << " test ACC(Viterbi): " << (double) tc_sum/tp_sum <<"   " << tc_sum << "/" << tp_sum << endl;

	string m_file = m_pModel->model_file; // model file path
	char buf[100];
	sprintf(buf,"%d",iteration);
	m_file+=buf;
	cout << m_file << endl;
	ofstream fout(m_file.c_str());
	fout << "num_params: " << m_pModel->num_params << endl;
	fout << "num_gates: " << m_pModel->num_gates << endl;
	fout << "window_size: " << m_pModel->window_size << endl;
	fout << "dim_features: " << m_pModel->dim_features << endl;
	fout << " weights: " << endl;
	for(int i=0;i<m_pModel->num_params;i++)
		fout << m_pModel->weights[i] << " ";
	fout << endl;
	fout.close();
}

	m_pModel->totalPos = m_pModel->totalCorrect = 0;
	for(int i=0;i<m_pModel->num_tst;i++){
		m_pModel->testData[i]->MAP();
		m_pModel->testData[i]->ComputeTestAccuracy();
	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalPos,&tp_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalCorrect,&tc_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
if(proc_id==0){
	cout << " test ACC(MAP): " << (double) tc_sum/tp_sum <<"   " << tc_sum << "/" << tp_sum << endl;
}
	m_pModel->totalPos = m_pModel->totalCorrect = 0;
	for(int i=0;i<m_pModel->num_data;i++){
		m_pModel->trainData[i]->ComputeGates();
		m_pModel->trainData[i]->ComputeVi();
		m_pModel->trainData[i]->ComputeViterbi();
		m_pModel->trainData[i]->ComputeTestAccuracy();
	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalPos,&tp_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Reduce(&m_pModel->totalCorrect,&tc_sum, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

if(proc_id==0)
	cout <<" train ACC: " << (double) tc_sum/tp_sum << "   " << tc_sum << "/" << tp_sum << endl;
}

void _LBFGS::ComputeGradient(vector<double>&g, const vector<double> &x, bool bCalculateGate)
{
	for(int i=0;i<m_pModel->num_params;i++) 
		m_pModel->weights[i] = x[i];
	//weights[num_params-1] = 1;
	double _norm = 0;
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Bcast(m_pModel->weights, m_pModel->num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);

	for(int i=0;i<m_pModel->num_params;i++) 
		m_pModel->grad[i]=0;
	for(int i=0;i<m_pModel->num_data;i++)
		m_pModel->trainData[i]->ComputeGradient(bCalculateGate);
	//for(int i=0;i<num_params;i++) g[i] = grad[i];
	
	for(int i=0;i<m_pModel->num_params;i++){
		if(i<m_pModel->num_states*(m_pModel->num_states+m_pModel->num_gates + 1))
			m_pModel->grad[i]=-m_pModel->grad[i]+m_pModel->weights[i]*m_pModel->reg[i]*2; //best:0.04
		else
			m_pModel->grad[i] = -m_pModel->grad[i]+m_pModel->weights[i]*m_pModel->reg[i]*2; //best:0.02
	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Reduce(m_pModel->grad, m_pModel->grad_sum, m_pModel->num_params, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Bcast(m_pModel->grad_sum, m_pModel->num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	for(int i=0;i<m_pModel->num_params;i++){
		g[i] = m_pModel->grad_sum[i];
		 _norm +=g[i]*g[i];
	}
	if(!proc_id)
		cerr << "Norm of Gradient: " << sqrt(_norm) << endl;
}

double _LBFGS::ComputeFunction(const vector<double> &x)
{
	for(int i=0;i<m_pModel->num_params;i++) 
		m_pModel->weights[i] = x[i];
	//weights[num_params-1] = 1;
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Bcast(m_pModel->weights, m_pModel->num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);

	double obj = 0, obj_sum = 0;
	for(int i=0;i<m_pModel->num_data;i++)
		obj += m_pModel->trainData[i]->Obj();
	obj = -obj;

	for(int i=0;i<x.size();i++){
		if(i<m_pModel->num_states*(m_pModel->num_states+m_pModel->num_gates+1))
			obj += m_pModel->weights[i]*m_pModel->weights[i]*m_pModel->reg[i];
		else 
			obj+=m_pModel->weights[i]*m_pModel->weights[i]*m_pModel->reg[i];
	}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Reduce(&obj, &obj_sum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
	MPI_Bcast(&obj_sum, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

	return obj_sum;
}

void _LBFGS::Report(const string &s)
{
	if(!proc_id) cerr << s << endl;
}

void bCNF_Model::SetSeed()
{
	unsigned int randomSeed=0;
	ifstream in("/dev/urandom",ios::in);
	in.read((char*)&randomSeed, sizeof(unsigned)/sizeof(char));
	in.close();
	        
	unsigned id=getpid();
	randomSeed=randomSeed*randomSeed+id*id;

	//we can set the random seed at only the main function
	srand48(randomSeed);
	srand(randomSeed);
}

int bypass;
double regularizer;
int num_test_set;

void bCNF_Model::SetParameters(int w_size, int n_states, int n_gates, int n_local)
{
	//setSeed();
	window_size = w_size; //best 13

	num_states = n_states;
	num_gates = n_gates; //best 20
	dim_one_pos = n_local;  // dim of local feature + 1 (for bias)

	dim_features = (window_size*dim_one_pos)+1;
	
	num_params = num_states*(1 + num_states + num_gates) + dim_features*num_gates;

	weights = new double[num_params];
	grad = new double[num_params];	
	grad_sum = new double[num_params];
	reg = new double[num_params];	
	double r=regularizer/num_procs; // regularization coefficients
	int i;
	for(i=0;i<num_states*(1+num_states);i++) 
		reg[i] = r;
	for(;i<num_states*(1+num_states+num_gates);i++) 
		reg[i] = r;
	for(;i<num_params;i++)
		reg[i] = r;
	
if(!proc_id){	
	cerr <<"num_params = " << num_params << endl;
	for(int i=0;i<num_params;i++)
			weights[i]=(drand48()-drand48())*0.1;
}
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Bcast(weights, num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Bcast(reg, num_params, MPI_DOUBLE, 0, MPI_COMM_WORLD);
}

void bCNF_Model::Initialize(string model_dir, int w_size, int n_states, int n_gates, int n_local, string output_file, string input_f)
{
	char fname[100];
	sprintf(fname, "/model.");
	model_file = model_dir+fname;
	SetParameters(w_size, n_states, n_gates, n_local);
	trainData.clear();
	testData.clear();
	LoadData(input_f);
	for(int i=0;i<num_data;i++)
		trainData[i]->makeFeatures();
	for(int i=0;i<num_tst;i++)
		testData[i]->makeFeatures();
}

void bCNF_Model::LoadData(string input)
{
	ifstream trn_in(input.c_str());
	trn_in >> num_data;
	if(!proc_id){
		cerr << "Load Data"<< dim_one_pos << ";" << num_data << endl;
	}
	int total_data = num_data;
	vector<SEQUENCE*> DATA;
	int length_s;
	double tmp;
	for(int i=0;i<num_data;i++){
		// construct a new sequence
		trn_in >> length_s;
		//if(!proc_id)
		//	cerr << length_s;
		
		SEQUENCE *seq = new SEQUENCE(length_s, this);

		for(int j=0;j<length_s;j++){
			for(int k=0;k<dim_one_pos-1;k++){
				trn_in >> seq->obs_feature[j][k];
			}

			seq->obs_feature[j][dim_one_pos-1] = 1;
		}
		for(int j=0;j<length_s;j++)
			trn_in >> seq->obs_label[j];
		if(i%num_procs != proc_id) {delete seq; continue;}
		DATA.push_back(seq);
	}
	trn_in.close();
	num_data = DATA.size();
	if (num_test_set>0)  //validation set?
		num_tst = num_test_set/num_procs;
	else //randomly sample a 1/5 validation set 
		num_tst = total_data/5/num_procs;

	        vector<int> shuffle;
        for(int i=0;i<num_data;i++)
                shuffle.push_back(num_data-1-i);
        SetSeed();
        random_shuffle(shuffle.begin(),shuffle.end());


	for(int i=0;i<num_tst;i++){
		testData.push_back(DATA[shuffle[i]]);
		//testData.push_back(DATA[num_data-i-1]);
	}
	num_data = num_data - num_tst;
	for(int i=0;i<num_data;i++){
		trainData.push_back(DATA[shuffle[i+num_tst]]);
		//trainData.push_back(DATA[i]);
		//cerr << i << " " << DATA[i]->length_seq << endl;
	}
}

void Usage()
{
	cerr << "CNF_mpi [-d <data dir>] [-w <window size>] [-s <number of states>] [-g <number of gates>] [-l <number of local features>] [-o output_file]\n";
}

int main(int argc, char **argv){

	MPI_Init(&argc, &argv);

	// the command line must be:
	if (argc < 3)
	{
		Usage();
		return 0;
	}

	num_test_set=0;
	bypass=0;
	regularizer=0.5;
	extern char* optarg;
	char c=0;
	string model_dir="./", output_file="cnf_train.dat";
	int w_size = 9; //best 13
	int n_states = 3, n_gates = 40; //best 20
	int n_local = 21;  // dim of local feature + 1 (for bias)
	string input_f="ss70";
	while((c=getopt(argc,argv,":w:s:g:d:o:i:l:b:r:f:t:"))!=EOF){
		switch(c)
		{
		case 'i':
			input_f=optarg;
			break;
		case 'w':
			w_size=atoi(optarg);
			break;
		case 's':
			n_states=atoi(optarg);
			break;
		case 'g':
			n_gates = atoi(optarg);
			break;
		case 'l':
			n_local = atoi(optarg);
			break;
		case 'o':
			output_file = optarg;
			break;
		case 'd':
			model_dir = optarg;
			if (model_dir[model_dir.size()-1] != '/')
				model_dir += "/";
			cerr << "model_dir=" << model_dir << endl;
			break;
		case 'r':
			regularizer=atof(optarg);
			break;
		case 't':
			num_test_set=atoi(optarg);
			break;
		default:
			Usage();
			exit(-1);
		}
	}

	bCNF_Model cnfModel;

	MPI_Comm_rank(MPI_COMM_WORLD, &proc_id);
	MPI_Comm_size(MPI_COMM_WORLD, &num_procs);

	cnfModel.Initialize(model_dir, w_size, n_states, n_gates, n_local, output_file, input_f);
	if(!proc_id){
		cerr << "Initialization Finished!" << endl;
	}

	_LBFGS* lbfgs = new _LBFGS(&cnfModel);
	vector<double> w_init(cnfModel.num_params,0);
	for(int i=0;i<cnfModel.num_params;i++)
		w_init[i] = cnfModel.weights[i];
	MPI_Barrier(MPI_COMM_WORLD);
	lbfgs->LBFGS(w_init, 2000);
	//lbfgs->Report(w_init, 1,1,1);
	MPI_Barrier(MPI_COMM_WORLD);
	for(int i=0;i<cnfModel.num_data;i++)
		delete cnfModel.trainData[i];
	for(int i=0;i<cnfModel.num_tst;i++)
		delete cnfModel.testData[i];
	MPI_Finalize();
	return 0;
}
